﻿Imports System

Namespace TTest
  Friend Class TTestProgram
	Shared Sub Main(ByVal args() As String)
            Console.WriteLine(vbCrLf & "Begin Welch's t-test using C# demo" & vbCrLf)
	  Dim x = New Double() { 88, 77, 78, 85, 90, 82, 88, 98, 90 }
	  Dim y = New Double() { 81, 72, 67, 81, 71, 70, 82, 81 }

            Console.WriteLine(vbCrLf & "The first data set (x) is:" & vbCrLf)
            ShowVector(x, 0)

            Console.WriteLine(vbCrLf & "The second data set (y) is:" & vbCrLf)
            ShowVector(y, 0)

            Console.WriteLine(vbCrLf & "Starting Welch's t-test using C#" & vbCrLf)
            TTest(x, y)

            Console.WriteLine(vbCrLf & "End t-test demo" & vbCrLf)
            Console.ReadLine()
        End Sub ' Main

        Private Shared Sub ShowVector(ByVal v() As Double, ByVal dec As Integer)
            For i As Integer = 0 To v.Length - 1
                Console.Write(v(i).ToString("F" & dec) & " ")
            Next i
            Console.WriteLine("")
        End Sub

        Public Shared Sub TTest(ByVal x() As Double, ByVal y() As Double)
            ' Welch's t-test for two samples
            Dim sumX As Double = 0.0 ' calculate means
            Dim sumY As Double = 0.0

            For i As Integer = 0 To x.Length - 1
                sumX += x(i)
            Next i

            For i As Integer = 0 To y.Length - 1
                sumY += y(i)
            Next i

            Dim n1 As Integer = x.Length
            Dim n2 As Integer = y.Length
            Dim meanX As Double = sumX / n1
            Dim meanY As Double = sumY / n2

            Dim sumXminusMeanSquared As Double = 0.0 ' calculate (sample) variances
            Dim sumYminusMeanSquared As Double = 0.0

            For i As Integer = 0 To n1 - 1
                sumXminusMeanSquared += (x(i) - meanX) * (x(i) - meanX)
            Next i

            For i As Integer = 0 To n2 - 1
                sumYminusMeanSquared += (y(i) - meanY) * (y(i) - meanY)
            Next i

            Dim varX As Double = sumXminusMeanSquared / (n1 - 1)
            Dim varY As Double = sumYminusMeanSquared / (n2 - 1)

            ' t statistic
            Dim top As Double = (meanX - meanY)
            Dim bot As Double = Math.Sqrt((varX / n1) + (varY / n2))
            Dim t As Double = top / bot


            ' df mildly tricky
            ' http://www.statsdirect.com/help/default.htm#parametric_methods/unpaired_t.htm
            Dim num As Double = ((varX / n1) + (varY / n2)) * ((varX / n1) + (varY / n2))
            Dim denomLeft As Double = ((varX / n1) * (varX / n1)) / (n1 - 1)
            Dim denomRight As Double = ((varY / n2) * (varY / n2)) / (n2 - 1)
            Dim denom As Double = denomLeft + denomRight
            Dim df As Double = num / denom

            Dim p As Double = Student(t, df) ' cumulative two-tail density

            Console.WriteLine("mean of x = " & meanX.ToString("F2"))
            Console.WriteLine("mean of y = " & meanY.ToString("F2"))
            Console.WriteLine("t = " & t.ToString("F4"))
            Console.WriteLine("df = " & df.ToString("F3"))
            Console.WriteLine("p-value = " & p.ToString("F5"))

            Explain()
        End Sub

        Public Shared Sub Explain()
            Console.Write(vbCrLf & "The p-value is the probability that ")
            Console.WriteLine("the two means are equal." & vbCrLf)
            Console.Write("If the p-value is low, typically less ")
            Console.WriteLine("than 0.05 or 0.01, then you")
            Console.Write("conclude there is statistcal evidence ")
            Console.WriteLine("the two means are different.")
            Console.Write("But if the p-value isn't below the 5% ")
            Console.WriteLine("or 1% critical value then you")
            Console.Write("conclude there's not enough evidence to ")
            Console.WriteLine("say the means are different." & vbCrLf)
        End Sub

	Public Shared Function Student(ByVal t As Double, ByVal df As Double) As Double
	  ' for large integer df or double df
	  ' adapted from ACM algorithm 395
	  ' returns 2-tail p-value
	  Dim n As Double = df ' to sync with ACM parameter name
	  Dim a, b, y As Double

	  t = t * t
	  y = t / n
	  b = y + 1.0
	  If y > 1.0E-6 Then
		  y = Math.Log(b)
	  End If
	  a = n - 0.5
	  b = 48.0 * a * a
	  y = a * y

	  y = (((((-0.4 * y - 3.3) * y - 24.0) * y - 85.5) / (0.8 * y * y + 100.0 + b) + y + 3.0) / b + 1.0) * Math.Sqrt(y)
	  Return 2.0 * Gauss(-y) ' ACM algorithm 209
	End Function

	Public Shared Function Student(ByVal t As Double, ByVal df As Integer) As Double
	  ' adapted from ACM algorithm 395
	  ' for small integer df, or int df with large t
	  Dim n As Integer = df ' to sync with ACM parameter name
	  Dim a, b, y, z As Double

	  z = 1.0
	  t = t * t
	  y = t / n
	  b = 1.0 + y

	  If n >= 20 AndAlso t < n OrElse n > 200 Then ' moderate df and smallish t ( df >= 20, t < df ) or very large df
		Dim x As Double = 1.0 * n ' make df a double
		Return Student(t, x) ' call double df version
	  End If

	  If n < 20 AndAlso t < 4.0 Then
		a = Math.Sqrt(y)
		y = Math.Sqrt(y)
		If n = 1 Then
		  a = 0.0
		End If
	  Else
		a = Math.Sqrt(b)
		y = a * n
		Dim j As Integer = 2
		Do While a <> z
		  z = a
		  y = y * (j - 1) / (b * j)
		  a = a + y / (n + j)
			j += 2
		Loop
		n = n + 2
		y = 0.0
		z = y
		a = -a
	  End If

	  Dim sanityCt As Integer = 0
	  Do While True AndAlso sanityCt < 10000
		sanityCt += 1
		n = n - 2
		If n > 1 Then
		  a = (n - 1) / (b * n) * a + y
		  Continue Do
		End If

		If n = 0 Then
		  a = a / Math.Sqrt(b)
		Else ' n == 1
		  a = (Math.Atan(y) + a / b) * 0.63661977236 ' 2/Pi
		End If

		Return z - a
	  Loop

	  Return -1.0 ' error
	End Function ' Student (int df)



	' ==========================================================

	'static double Gauss(double z)
	'{
	'  // aka Phi(z)
	'  // Abramowitz and Stegun eq. 7.1.26
	'  double p = 0.3275911;
	'  double a1 = 0.254829592;
	'  double a2 = -0.284496736;
	'  double a3 = 1.421413741;
	'  double a4 = -1.453152027;
	'  double a5 = 1.061405429;

	'  int sign;
	'  if (z < 0.0)
	'    sign = -1;
	'  else
	'    sign = 1;

	'  double x = Math.Abs(z) / Math.Sqrt(2.0);
	'  double t = 1.0 / (1.0 + p * x);
	'  double erf = 1.0 - (((((a5 * t + a4) * t) + a3) * t + a2) * t + a1) * t * Math.Exp(-x * x);
	'  return 0.5 * (1.0 + (sign * erf));
	'}

	Public Shared Function Gauss(ByVal z As Double) As Double
	  ' input = z-value (-inf to +inf)
	  ' output = p under Standard Normal curve from -inf to z
	  ' e.g., if z = 0.0, function returns 0.5000
	  ' ACM Algorithm #209
	  Dim y As Double ' 209 scratch variable
	  Dim p As Double ' result. called 'z' in 209
	  Dim w As Double ' 209 scratch variable

	  If z = 0.0 Then
		p = 0.0
	  Else
		y = Math.Abs(z) / 2
		If y >= 3.0 Then
		  p = 1.0
		ElseIf y < 1.0 Then
		  w = y * y
		  p = ((((((((0.000124818987 * w - 0.001075204047) * w + 0.005198775019) * w - 0.019198292004) * w + 0.059054035642) * w - 0.151968751364) * w + 0.319152932694) * w - 0.531923007300) * w + 0.797884560593) * y * 2.0
		Else
		  y = y - 2.0
		  p = (((((((((((((-0.000045255659 * y + 0.000152529290) * y - 0.000019538132) * y - 0.000676904986) * y + 0.001390604284) * y - 0.000794620820) * y - 0.002034254874) * y + 0.006549791214) * y - 0.010557625006) * y + 0.011630447319) * y - 0.009279453341) * y + 0.005353579108) * y - 0.002141268741) * y + 0.000535310849) * y + 0.999936657524
		End If
	  End If

	  If z > 0.0 Then
		Return (p + 1.0) / 2
	  Else
		Return (1.0 - p) / 2
	  End If
	End Function ' Gauss()

  End Class ' Program

End Namespace ' ns
